/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: turbo_channel.c,v $
 *
 * HISTORY
 * $Log: turbo_channel.c,v $
 * Revision 1.3  1996/01/30  22:54:40  slm
 * Updated copyright and version.
 *
 * Revision 1.2.1.1  1994/12/02  23:57:04  hkaram
 * New branch
 *
 * Revision 1.2  1994/06/11  00:02:48  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/05/14  23:36:07  davidm
 * Initial revision
 */
/*
 * TURBOchannel interface.  Determines what options are installed and
 * fills in the appriopriate bus descriptors.
 */
#include "scout_device.h"
#include "scout_trace.h"
#include "scout_util.h"
#include "alpha_misc.h"
#include "turbo_channel.h"

int trace_turbo_channel = 0;

#define MSEC	1000	/* micro-seconds per milli-second */


/*
 * The TURBOchannel standard defines the first byte of the ROM to be
 * the ROM width.  ROM stride is defined to be 4.  Thus, by reading
 * the first byte, we have all the info we need on the ROM layout.  To
 * be safe on the safe side , we do not copy anything unless the ROM
 * width is no more than the stride (i.e., 4).
 */
static inline void
copy_rom_header(u_char *from, struct tc_rom_header *tcrh)
{
    int i, width, nbytes = sizeof(*tcrh);
    u_char *to = (u_char*) tcrh;

    width = *from;		/* first byte is width */
    if (width <= TC_ROM_STRIDE) {
	i = 0;
	while (nbytes-- > 0) {
	    *to++ = from[i++];
	    if (i >= width) {
		i = 0;
		from += TC_ROM_STRIDE;
	    } /* if */
	} /* while */
    } else {
	tcrh->rom_signature = ~TC_ROM_SIGNATURE;	/* invalidate entry */
    } /* if */
} /* copy_rom_header */


void
tc_config(void *base, long slot_size, int nslots, Scout_IPL tc_ipl)
{
    int i, j, k;
    char *slot_base;
    struct tc_rom_header tcrh;
    Scout_Device_Info di;
    char name[sizeof(tcrh.module_name) + 1];
    int off[] = {0x3c0000 + TC_OFF_ROM_HEADER, TC_OFF_ROM_HEADER};

    di.name = name;

    scout_udelay(200*MSEC);	/* must not access TC for the first 200 ms */

    slot_base = ((char*)base + (nslots - 1) * slot_size);

    for (i = nslots - 1; i >= 0; --i) {
	for (j = 0; j < sizeof(off)/sizeof(off[0]); ++j) {
	    if (!probe_read(slot_base + off[j])) {
		ScoutTrace(turbo_channel, TR_DETAILED,
			   ("tc%d: no access to addr %p",
			    i, slot_base + off[j]));
		continue;
	    } /* if */
	    copy_rom_header(slot_base + off[j], &tcrh);
	    if (tcrh.rom_signature == TC_ROM_SIGNATURE &&
		tcrh.rom_stride == TC_ROM_STRIDE)
	    {
		ScoutTrace(turbo_channel, TR_DETAILED,
			   ("tc%d: signature = %#x, width=%d, stride=%d, "
			    "size=%d, min slot size=%d, flags=%x",
			    i, tcrh.rom_signature, tcrh.rom_width,
			    tcrh.rom_stride, tcrh.rom_size, tcrh.slot_size,
			    tcrh.flags));
		switch (tcrh.rom_width) {
		  case 1: case 2: case 4: break;
		  default:
		    printf("tc%d: error: option ROM has bad width %d\n",
			   i, tcrh.rom_width);
		    continue;
		} /* switch */
		if (tcrh.slot_size > (slot_size >> 22)) {
		    printf("tc%d: error: option requires slot size of %d MB\n",
			   i, 4 * tcrh.slot_size);
		    continue;
		} /* if */

		printf("tc%d: %.8s %.8s %.8s %.4s at 0x%p\n", i,
		       tcrh.module_name, tcrh.vendor, tcrh.firmware_version,
		       tcrh.firmware_type, slot_base + off[j]);

		for (k = 0; k < sizeof(tcrh.module_name); ++k) {
		    if (tcrh.module_name[k] == ' ') {
			break;
		    } /* if */
		    di.name[k] = tcrh.module_name[k];
		} /* while */
		di.name[k] = '\0';
		di.base_addr = P_TO_K0SEG(slot_base);
		di.ipl = tc_ipl;
		di.rom_offset = off[j];
		scout_device_add(&di);
		break;
	    } /* if */
	} /* for */
	slot_base -= slot_size;
    } /* for */
} /* tc_config */

			/*** end of turbo_channel.c ***/
