/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_sync_arch.h,v $
 *
 * HISTORY
 * $Log: scout_sync_arch.h,v $
 * Revision 1.1  1996/01/30  22:55:35  slm
 * Initial revision
 *
 * Revision 1.5.1.1  1994/12/03  00:00:31  hkaram
 * New branch
 *
 * Revision 1.5  1994/11/20  01:34:20  davidm
 * Basic-block profiling can make atomic sequences long.  Up
 * SCOUT_SYNC_MAX_LENGTH in that case.
 *
 * Revision 1.4  1994/06/11  00:02:54  menze
 * Added copyright notice
 *
 * Revision 1.3  1994/04/30  01:25:01  davidm
 * "scout.h" should not be included for assembly code.
 *
 * Revision 1.2  1994/04/30  01:00:58  davidm
 * Reorganized somewhat.
 *
 * Revision 1.1  1994/04/14  23:45:14  davidm
 * Initial revision
 */
#ifndef scout_sync_arch_h
#define scout_sync_arch_h

#include "scout_config.h"

#define SCOUT_MAX_IPL		7	/* max. interrupt priority */

#ifdef SCOUT_ROLLFORWARD
# ifdef SCOUT_BBPROFILE
#  define SCOUT_SYNC_MAX_LENGTH		(64*4)	/* max length in bytes */
# else
#  define SCOUT_SYNC_MAX_LENGTH		(32*4)	/* max length in bytes */
# endif
# define SCOUT_SYNC_LANDMARK		0x68210000
# define SCOUT_SYNC_TARGET_REG		 1	/* reg holding jmp dest addr */
# define SCOUT_SYNC_RET_ADDR_REG	26	/* reg holding return PC */
#endif /* SCOUT_ROLLFORWARD */

#ifndef LANGUAGE_ASSEMBLY

# include "scout.h"
# include "scout_assert.h"
# include "alpha_pal.h"

  typedef int Scout_IPL;

# define scout_swap_ipl(l) \
({ \
    register u_long v0 asm ("0"); \
    register u_long a0 asm ("16"); \
    a0 = (l); \
    asm volatile ("call_pal %2 # %0 %1" : "r="(v0), "r="(a0) \
		  : "i"(PAL_swpipl), "r"(a0) \
		  : "memory", "0", "1", "16", "22", "23", "24", "25"); \
    v0; \
})

# ifdef SCOUT_ROLLFORWARD

#  define ScoutAtomicSequence(as) \
{ \
   __label__ the_end, the_start; \
   register void *end_pc asm ("$1"); \
   scout_assert(&&the_end - &&the_start < SCOUT_SYNC_MAX_LENGTH); \
   asm volatile ("" ::: "memory"); \
   end_pc = &&the_end; \
   asm volatile ("" ::: "memory"); \
 the_start: \
   {as;} \
   asm volatile ("jmp %0,(%1),0" : "r="(end_pc) : "0"(end_pc) : "memory"); \
   asm volatile ("" ::: "memory"); \
 the_end: \
}

# else /* SCOUT_ROLLFORWARD */

#  define ScoutAtomicSequence(as) \
{ \
   Scout_IPL old_ipl = scout_swap_ipl(SCOUT_MAX_IPL); \
   {as;} \
   scout_swap_ipl(old_ipl); \
}

# endif /* SCOUT_ROLLFORWARD */
#endif /* LANGUAGE_ASSEMBLY */

#endif /* scout_sync_arch_h */
